from types import SimpleNamespace

ADATA = SimpleNamespace(
    LAYERS="layers",
    OBS="obs",
    OBSM="obsm",
    OBSP="obsp",
    VAR="var",
    VARM="varm",
    VARP="varp",
    X="X",
)

ADATA_DICTLIKE_ATTRS = {ADATA.LAYERS, ADATA.OBSM, ADATA.OBSP, ADATA.VARM, ADATA.VARP}
ADATA_IDENTITY_ATTRS = {ADATA.OBS, ADATA.VAR, ADATA.X}
ADATA_FEATURELEVEL_ATTRS = {ADATA.LAYERS, ADATA.VAR, ADATA.VARM, ADATA.X}
ADATA_ARRAY_ATTRS = {
    ADATA.LAYERS,
    ADATA.OBSM,
    ADATA.OBSP,
    ADATA.VARM,
    ADATA.VARP,
    ADATA.X,
}

SCVERSE_DEFS = SimpleNamespace(ADATA="adata", MDATA="mdata")

VALID_MUDATA_LEVELS = {SCVERSE_DEFS.ADATA, SCVERSE_DEFS.MDATA}
